@@ -4,6 +4,7 @@
 #include "libfilezilla.hpp"
 
 #include <algorithm>
+#include <cstdint>
 #include <string>
 #include <string_view>
 #include <vector>
@@ -124,9 +125,9 @@ struct FZ_PUBLIC_SYMBOL less_insensitive_ascii final
 	template<typename T>
 	bool operator()(T const& lhs, T const& rhs) const {
 		return std::lexicographical_compare(lhs.cbegin(), lhs.cend(), rhs.cbegin(), rhs.cend(),
-		    [](typename T::value_type const& a, typename T::value_type const& b) {
-			    return tolower_ascii(a) < tolower_ascii(b);
-		    }
+			[](typename T::value_type const& a, typename T::value_type const& b) {
+				return tolower_ascii(a) < tolower_ascii(b);
+			}
 		);
 	}
 };
@@ -138,17 +139,17 @@ struct FZ_PUBLIC_SYMBOL less_insensitive_ascii final
 inline bool equal_insensitive_ascii(std::string_view a, std::string_view b)
 {
 	return std::equal(a.cbegin(), a.cend(), b.cbegin(), b.cend(),
-	    [](auto const& a, auto const& b) {
-		    return tolower_ascii(a) == tolower_ascii(b);
-	    }
+		[](auto const& a, auto const& b) {
+			return tolower_ascii(a) == tolower_ascii(b);
+		}
 	);
 }
 inline bool equal_insensitive_ascii(std::wstring_view a, std::wstring_view b)
 {
 	return std::equal(a.cbegin(), a.cend(), b.cbegin(), b.cend(),
-	    [](auto const& a, auto const& b) {
-		    return tolower_ascii(a) == tolower_ascii(b);
-	    }
+		[](auto const& a, auto const& b) {
+			return tolower_ascii(a) == tolower_ascii(b);
+		}
 	);
 }
 
@@ -157,20 +158,20 @@ inline bool equal_insensitive_ascii(std::wstring_view a, std::wstring_view b)
  * \return the converted string on success. On failure an empty string is returned.
  */
 std::wstring FZ_PUBLIC_SYMBOL to_wstring(std::string_view const& in);
-inline std::wstring FZ_PUBLIC_SYMBOL to_wstring(std::wstring_view const& in) { return std::wstring(in); }
 
 /** \brief Returns identity, that way to_wstring can be called with native_string.
  *
- * This template prevents converting std::wstring to std::wstring_view and back to std::wstring
+ * This template deals with wide string literals, std::wstring and std::wstring_view parameters.
  */
-template<typename T, typename std::enable_if_t<std::is_same_v<std::wstring, typename std::decay_t<T>>, int> = 0>
-inline std::wstring to_wstring(T const& in) {
-	return in;
+template <typename T>
+inline auto to_wstring(T && in) -> decltype(std::wstring(std::forward<T>(in)))
+{
+	return std::wstring(std::forward<T>(in));
 }
 
 /// Converts from arithmetic type to std::wstring
 template<typename Arg>
-inline typename std::enable_if<std::is_arithmetic<std::decay_t<Arg>>::value, std::wstring>::type to_wstring(Arg && arg)
+inline typename std::enable_if<std::is_arithmetic_v<std::decay_t<Arg>>, std::wstring>::type to_wstring(Arg && arg)
 {
 	return std::to_wstring(std::forward<Arg>(arg));
 }
@@ -191,21 +192,21 @@ std::wstring FZ_PUBLIC_SYMBOL to_wstring_from_utf8(fz::buffer const& in);
  * \return the converted string on success. On failure an empty string is returned.
  */
 std::string FZ_PUBLIC_SYMBOL to_string(std::wstring_view const& in);
-inline std::string FZ_PUBLIC_SYMBOL to_string(std::string_view const& in) { return std::string(in); }
 
 /** \brief Returns identity, that way to_wstring can be called with native_string.
  *
- * This template prevents converting std::wstring to std::wstring_view and back to std::wstring
+ * This template deals with string literals, std::string and std::string_view parameters.
  */
-template<typename T, typename std::enable_if_t<std::is_same_v<std::string, typename std::decay_t<T>>, int> = 0>
-inline std::string to_string(T const& in) {
-	return in;
+template <typename T>
+inline auto to_string(T && in) -> decltype(std::string(std::forward<T>(in)))
+{
+	return std::string(std::forward<T>(in));
 }
 
 
 /// Converts from arithmetic type to std::string
 template<typename Arg>
-inline typename std::enable_if<std::is_arithmetic<std::decay_t<Arg>>::value, std::string>::type to_string(Arg && arg)
+inline typename std::enable_if<std::is_arithmetic_v<std::decay_t<Arg>>, std::string>::type to_string(Arg && arg)
 {
 	return std::to_string(std::forward<Arg>(arg));
 }
@@ -236,13 +237,13 @@ std::string FZ_PUBLIC_SYMBOL to_utf8(std::wstring_view const& in);
 
 /// Calls either fz::to_string or fz::to_wstring depending on the passed template argument
 template<typename String, typename Arg>
-inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same<String, std::string>::value, decltype(to_string(std::forward<Arg>(arg)))>::type
+inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same_v<String, std::string>, decltype(to_string(std::forward<Arg>(arg)))>::type
 {
 	return to_string(std::forward<Arg>(arg));
 }
 
 template<typename String, typename Arg>
-inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same<String, std::wstring>::value, decltype(to_wstring(std::forward<Arg>(arg)))>::type
+inline auto toString(Arg&& arg) -> typename std::enable_if<std::is_same_v<String, std::wstring>, decltype(to_wstring(std::forward<Arg>(arg)))>::type
 {
 	return to_wstring(std::forward<Arg>(arg));
 }
@@ -285,18 +286,183 @@ template<> inline wchar_t const* choose_string(char const*, wchar_t const* w) {
 #define fzS(Char, s) fz::choose_string<Char>(s, L ## s)
 #endif
 
- /// Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
+ /** \brief Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
+  *
+  * \arg find If empty, no replacement takes place.
+  */
 std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string_view const& in, std::string_view const& find, std::string_view const& replacement);
 std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring_view const& in, std::wstring_view const& find, std::wstring_view const& replacement);
+
+/// Returns \c in with all occurrences of \c find in the input string replaced with \c replacement
 std::string FZ_PUBLIC_SYMBOL replaced_substrings(std::string_view const& in, char find, char replacement);
 std::wstring FZ_PUBLIC_SYMBOL replaced_substrings(std::wstring_view const& in, wchar_t find, wchar_t replacement);
 
-/// Modifies \c in, replacing all occurrences of \c find with \c replacement
+/** \brief Modifies \c in, replacing all occurrences of \c find with \c replacement
+ *
+ * \arg find If empty, no replacement takes place.
+ */
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::string& in, std::string_view const& find, std::string_view const& replacement);
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::wstring& in, std::wstring_view const& find, std::wstring_view const& replacement);
+
+/// Modifies \c in, replacing all occurrences of \c find with \c replacement
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::string& in, char find, char replacement);
 bool FZ_PUBLIC_SYMBOL replace_substrings(std::wstring& in, wchar_t find, wchar_t replacement);
 
+/**
+ * \brief Container-like class that can be used to iterate over tokens in a string.
+ *
+ * The class will keep a copy of any temporary constructor's parameter.
+ * strtokenizer must live longer than the iterators created from it.
+ *
+ * Access through the iterators returns views, make sure non-temporary arguments
+ * to strtokenizer constructur live longer than the iterators.
+ *
+ * Do not modify a string for which there is an strtokenizer instance.
+ *
+ * Always use the construction guide for this class, never explictly use the
+ * template parameters.
+ *
+ * Usage example:
+ * \code
+ * for (auto t : fz::strtokenizer("foo,baz,,bar", ",", true)) {
+ *   std::cout << t << "\n";
+ * }
+ *
+ * // Will print the following:
+ * //     foo
+ * //     baz
+ * //     bar
+ * \endcode
+ */
+template <typename String, typename Delims>
+class strtokenizer
+{
+	using view_type = std::basic_string_view<std::decay_t<decltype(std::declval<String>()[0])>>;
+
+public:
+	/**
+	 * \brief strtokenizer class constructor.
+	 *
+	 * \param delims the delimiters to look for
+	 * \param ignore_empty If true, empty tokens are omitted in the output
+	 */
+	constexpr strtokenizer(String && string, Delims &&delims, bool ignore_empty)
+		: string_(std::forward<String>(string))
+		, delims_(std::forward<Delims>(delims))
+		, ignore_empty_(ignore_empty)
+	{}
+
+	using value_type = const view_type;
+	using pointer = value_type*;
+	using reference = value_type&;
+	using size_type = std::size_t;
+	using difference_type = std::ptrdiff_t;
+
+	struct sentinel{};
+
+	struct iterator
+	{
+		using iterator_category = std::input_iterator_tag;
+		using difference_type   = strtokenizer::difference_type;
+		using value_type        = strtokenizer::value_type;
+		using pointer           = strtokenizer::pointer;
+		using reference         = strtokenizer::reference;
+
+		constexpr bool operator !=(sentinel) const
+		{
+			return !s_.empty();
+		}
+
+		constexpr bool operator ==(sentinel) const
+		{
+			return s_.empty();
+		}
+
+		constexpr value_type operator*() const
+		{
+			return s_.substr(0, pos_);
+		}
+
+		constexpr iterator &operator++()
+		{
+			for (;;) {
+				if (pos_ != s_.size()) {
+					++pos_;
+				}
+
+				s_.remove_prefix(pos_);
+
+				pos_ = s_.find_first_of(t_->delims_);
+
+				if (pos_ == view_type::npos) {
+					pos_ = s_.size();
+					break;
+				}
+
+				if (pos_ != 0 || !t_->ignore_empty_) {
+					break;
+				}
+			}
+
+			return *this;
+		}
+
+	private:
+		friend strtokenizer;
+
+		constexpr iterator(const strtokenizer *t)
+			: t_(t)
+			, s_(view_type(t_->string_))
+			, pos_(view_type::npos)
+		{
+			operator++();
+		}
+
+		const strtokenizer *t_;
+		view_type s_;
+		size_type pos_;
+	};
+
+	using const_value_type = value_type;
+	using const_pointer = pointer;
+	using const_reference = reference;
+	using const_iterator = iterator;
+
+	constexpr iterator begin() const
+	{
+		return { this };
+	}
+
+	constexpr sentinel end() const
+	{
+		return {};
+	}
+
+	constexpr const_iterator cbegin() const
+	{
+		return { this };
+	}
+
+	constexpr sentinel cend() const
+	{
+		return {};
+	}
+
+public:
+	String string_;
+	Delims delims_;
+	bool ignore_empty_;
+};
+
+/**
+ * \brief strtokenizer class construction-guide.
+ *
+ * \param delims the delimiters to look for
+ * \param ignore_empty If true, empty tokens are omitted in the output
+ */
+template <typename String, typename Delims>
+strtokenizer(String && string, Delims &&delims, bool ignore_empty) -> strtokenizer<String, Delims>;
+
 /**
  * \brief Tokenizes string.
  *
@@ -333,7 +499,10 @@ inline auto FZ_PUBLIC_SYMBOL strtok_view(std::wstring_view const& tokens, wchar_
 template<typename T, typename String>
 T to_integral_impl(String const& s, T const errorval = T())
 {
-	if constexpr (std::is_enum_v<T>) {
+	if constexpr (std::is_same_v<T, bool>) {
+		return static_cast<T>(to_integral_impl<unsigned int>(s, static_cast<unsigned int>(errorval))) != 0;
+	}
+	else if constexpr (std::is_enum_v<T>) {
 		return static_cast<T>(to_integral_impl<std::underlying_type_t<T>>(s, static_cast<std::underlying_type_t<T>>(errorval)));
 	}
 	else {
@@ -535,6 +704,63 @@ bool ends_with(String const& s, String const& ending)
 std::string FZ_PUBLIC_SYMBOL normalize_hyphens(std::string_view const& in);
 std::wstring FZ_PUBLIC_SYMBOL normalize_hyphens(std::wstring_view const& in);
 
+/// Verifies that the input data is valid UTF-8.
+bool FZ_PUBLIC_SYMBOL is_valid_utf8(std::string_view s);
+
+/**
+ * \brief Verifies that the input data is valid UTF-8.
+ *
+ * The verfication state is exposed, you can use this function to verify
+ * data to be in UTF-8 in a piecewise fashion.
+ * When starting verification, initialize state with 0 and call the function
+ * for as many blocks of data as needed, each time passing the previously
+ * updated state along.
+ *
+ * If the input data on any particular call ends in the midlde of a UTF-8
+ * sequence, state is updated, so that the the check can continue with the
+ * next block of data.
+ *
+ * Once you have no data left to verify, check that the state is zero. If it
+ * is non-zero, the input data was prematurely terminated in the middle of a
+ * UTF-8 sequence.
+ *
+ * If the input data is invalid, the function returns false and state is
+ * updated with the offset of the offending input byte.
+ */
+bool FZ_PUBLIC_SYMBOL is_valid_utf8(std::string_view s, size_t & state);
+
+/**
+ * \brief Encodes a valid Unicode codepoint as UTF-8 and appends it to the passed string
+ *
+ * Undefined if not passed a valid codepoint.
+ */
+void FZ_PUBLIC_SYMBOL unicode_codepoint_to_utf8_append(std::string& result, uint32_t codepoint);
+
+/**
+ * \brief Converts from UTF-16-BE and appends it to the passed string
+ *
+ * The conversion state is exposed, you can use this function to convert
+ * data to UTF-8 in a piecewise fashion.
+ * When starting conversion, initialize state with 0 and call the function
+ * for as many blocks of data as needed, each time passing the previously
+ * updated state along.
+ *
+ * If the input data on any particular call ends in the midlde of a UTF-16
+ * sequence, state is updated, so that the conversion can continue at the next
+ * iteration.
+ *
+ * Once you have no data left to convert, check that the state is zero. If it
+ * is non-zero, the input data was prematurely terminated in the middle of a
+ * UTF-16 sequence.
+ *
+ * If the input data is invalid, the function returns false and state is
+ * updated with the offset of the offending input byte.
+ */
+bool FZ_PUBLIC_SYMBOL utf16be_to_utf8_append(std::string & result, std::string_view data, uint32_t & state);
+
+/// Just as utf16be_to_utf8_append but for little-endian UTF-16.
+bool FZ_PUBLIC_SYMBOL utf16le_to_utf8_append(std::string & result, std::string_view data, uint32_t & state);
+
 }
 
 #endif
