@@ -215,6 +215,8 @@ public:
 	/** \brief Get timestamp as struct tm
 	 *
 	 * Undefined if datetime is empty.
+	 * 
+	 * Note: On Windows the tm_yday and tm_isdst fields are undefined.
 	 */
 	tm get_tm(zone z) const;
 
@@ -415,6 +417,10 @@ public:
 	{
 		return monotonic_clock(*this) += d;
 	}
+	monotonic_clock const operator-(duration const& d) const
+	{
+		return monotonic_clock(*this) += d;
+	}
 
 private:
 	typedef std::chrono::steady_clock clock_type;
